#!/usr/bin/env python3

import sys
import click
import traceback
from functools import update_wrapper
from cnaas_sync.netbox import NetBoxAPI
from cnaas_sync.utils import print_color
from cnaas_sync.ni import NI as ni_server
from cnaas_sync.nav import NAV as nav_server
from cnaas_sync.nms import NMS as nms_server
from importlib.metadata import PackageNotFoundError, version as importlib_version
from cnaas_sync.settings import Settings, ConfigSettingsBuilder, ConfigManager


def cli_flags(func):
    @click.option(
        "--config",
        required=False,
        type=click.Path(exists=True),
        help="Path to the configuration file.",
    )
    @click.option(
        "--dry-run",
        is_flag=True,
        required=False,
        help="Simulate the sync actions without executing them.",
    )
    @click.option(
        "--show-settings",
        is_flag=True,
        required=False,
        help="Show the environment variables been used.",
    )
    @click.option(
        "--verbose", is_flag=True, required=False, help="Enable verbose output."
    )
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return update_wrapper(wrapper, func)


def nms_credentials(func):
    @click.option(
        "--nms-base-url", required=False, default=None, help="Base URL for NMS."
    )
    @click.option(
        "--nms-client-id", required=False, default=None, help="Client ID for NMS."
    )
    @click.option(
        "--nms-client-secret",
        required=False,
        default=None,
        help="Client secret for NMS.",
    )
    @click.option(
        "--nms-client-endpoint",
        required=False,
        default=None,
        help="Client endpoint for NMS.",
    )
    @cli_flags
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return update_wrapper(wrapper, func)


def build_nav_config(
    nav_base_url,
    nav_api_key,
    ni_base_url,
    ni_user_name,
    ni_api_key,
    nms_base_url,
    nms_client_id,
    nms_client_secret,
    nms_client_endpoint,
):
    # Initialize ConfigSettingsBuilder
    config_builder = ConfigSettingsBuilder()

    # Add NAV section and set its values
    config_builder.add_section("NAV")
    config_builder.set_value("NAV", "NAV_BASE_URL", nav_base_url)
    config_builder.set_value("NAV", "NAV_API_KEY", nav_api_key)

    # Add NI section and set its values
    config_builder.add_section("NI")
    config_builder.set_value("NI", "NI_BASE_URL", ni_base_url)
    config_builder.set_value("NI", "NI_USERNAME", ni_user_name)
    config_builder.set_value("NI", "NI_USER_API_KEY", ni_api_key)

    # Add NMS section and set its values
    config_builder.add_section("NMS")
    config_builder.set_value("NMS", "NMS_BASE_URL", nms_base_url)
    config_builder.set_value("NMS", "NMS_OAUTH_CLIENT_ID", nms_client_id)
    config_builder.set_value("NMS", "NMS_OAUTH_CLIENT_SECRET", nms_client_secret)
    config_builder.set_value("NMS", "NMS_OAUTH_TOKEN_ENDPOINT", nms_client_endpoint)

    # Return the config object
    return config_builder.get_config()


def build_ni_config(
    ni_base_url,
    ni_user_name,
    ni_api_key,
    nms_base_url,
    nms_client_id,
    nms_client_secret,
    nms_client_endpoint,
):
    # Initialize ConfigSettingsBuilder
    config_builder = ConfigSettingsBuilder()

    # Add NI section and set its values
    config_builder.add_section("NI")
    config_builder.set_value("NI", "NI_BASE_URL", ni_base_url)
    config_builder.set_value("NI", "NI_USERNAME", ni_user_name)
    config_builder.set_value("NI", "NI_USER_API_KEY", ni_api_key)

    # Add NMS section and set its values
    config_builder.add_section("NMS")
    config_builder.set_value("NMS", "NMS_BASE_URL", nms_base_url)
    config_builder.set_value("NMS", "NMS_OAUTH_CLIENT_ID", nms_client_id)
    config_builder.set_value("NMS", "NMS_OAUTH_CLIENT_SECRET", nms_client_secret)
    config_builder.set_value("NMS", "NMS_OAUTH_TOKEN_ENDPOINT", nms_client_endpoint)

    # Return the config object
    return config_builder.get_config()


def update_ni(dry_run: str, verbose: str, settings: Settings):
    nms = nms_server(
        nms_client_id=settings.nms.oauth_client_id,
        nms_client_secret=settings.nms.oauth_client_secret,
        nms_token_endpoint=settings.nms.oauth_token_endpoint,
        urlbase=settings.nms.base_url,
        dry_run=dry_run,
    )
    nms_devices = nms.get_all_devices()
    ni = ni_server(
        settings.ni.base_url,
        settings.ni.username,
        settings.ni.user_api_key,
        verbose=verbose,
    )
    ni.update(nms_devices)


def update_nav_with_ni_nms(dry_run: str, verbose: str, settings: Settings):
    nms = nms_server(
        nms_client_id=settings.nms.oauth_client_id,
        nms_client_secret=settings.nms.oauth_client_secret,
        nms_token_endpoint=settings.nms.oauth_token_endpoint,
        urlbase=settings.nms.base_url,
        dry_run=dry_run,
    )

    ni = ni_server(
        settings.ni.base_url,
        settings.ni.username,
        settings.ni.user_api_key,
        verbose=verbose,
    )
    nav = nav_server(settings.nav.base_url, settings.nav.api_key, verbose=verbose)

    ni_devices = ni.get_all_devices()
    nav.update(ni_devices, nms)


@click.group(invoke_without_command=True)
@click.option("--version", is_flag=True, help="Show the version of cnaas-sync.")
def cli(version):
    """CLI tool for syncing and listing devices."""
    if version:
        try:
            package_version = importlib_version("cnaas-sync")
            click.echo(f"cnaas-sync:{package_version}")
        except PackageNotFoundError:
            click.echo("Package cnaas-sync is not installed.")


@click.command()
@click.option(
    "--ni-base-url", required=False, default=None, help="Base URL for NI API."
)
@click.option("--ni-user-name", required=False, default=None, help="User name for NI.")
@click.option("--ni-api-key", required=False, default=None, help="API key for NI.")
@click.option(
    "--source",
    default="nms",
    required=False,
    show_default=True,
    help="Specify the source system (e.g., nms).",
)
@nms_credentials
def ni(
    config,
    source,
    ni_base_url,
    ni_user_name,
    ni_api_key,
    nms_base_url,
    nms_client_id,
    nms_client_secret,
    nms_client_endpoint,
    dry_run,
    show_settings,
    verbose,
):
    """
    Execute synchronization for the NI service from NMS.
    """
    config_dict = {
        "ni_base_url": ni_base_url,
        "ni_user_name": ni_user_name,
        "ni_api_key": ni_api_key,
        "nms_base_url": nms_base_url,
        "nms_client_id": nms_client_id,
        "nms_client_secret": nms_client_secret,
        "nms_client_endpoint": nms_client_endpoint,
    }

    config_manager = ConfigManager(
        build_function=build_ni_config, config=config, op_parameters=config_dict
    )
    settings = config_manager.check_config()

    if show_settings:
        print_color(settings, color="blue")

    update_ni(dry_run, verbose, settings)


@click.command()
@click.option("--netbox-url", required=False, default=None, help="Base URL for Netbox.")
@click.option(
    "--netbox-api-token",
    required=False,
    default=None,
    help="API token/key for Netbox.",
)
@click.option(
    "--netbox-tenant-name",
    required=False,
    default=None,
    help="Tenant name to be synced in Netbox.",
)
@click.option(
    "--nav-base-url", required=False, default=None, help="Base URL for NAV API."
)
@click.option("--nav-api-key", required=False, default=None, help="API key for NAV.")
@click.option(
    "--ni-base-url", required=False, default=None, help="Base URL for NI API."
)
@click.option("--ni-user-name", required=False, default=None, help="User name for NI.")
@click.option("--ni-api-key", required=False, default=None, help="API key for NI.")
@click.option(
    "--source",
    type=click.Choice(["nms", "netbox"]),
    default="nms",
    required=True,
    show_default=True,
    help="Specify the source system (e.g., nms or netbox).",
)
@click.option(
    "--nav-organization",
    default="SUNET",
    required=False,
    show_default=True,
    help="Specify the organization to be used for syncing. Currently the API does not support dynamic org creation",
)
@nms_credentials
def nav(
    config,
    source,
    nav_organization,
    netbox_url,
    netbox_api_token,
    netbox_tenant_name,
    nav_base_url,
    nav_api_key,
    ni_base_url,
    ni_user_name,
    ni_api_key,
    nms_base_url,
    nms_client_id,
    nms_client_secret,
    nms_client_endpoint,
    dry_run,
    show_settings,
    verbose,
):
    """
    Execute synchronization for the Nav service from NMS.
    """
    config_dict = {
        "netbox_url": netbox_url,
        "netbox_api_token": netbox_api_token,
        "netbox_tenant_name": netbox_tenant_name,
        "nav_base_url": nav_base_url,
        "nav_api_key": nav_api_key,
        "nms_base_url": nms_base_url,
        "nms_client_id": nms_client_id,
        "nms_client_secret": nms_client_secret,
        "nms_client_endpoint": nms_client_endpoint
    }

    required_envs = [
        "NETBOX_URL",
        "NETBOX_API_TOKEN",
        "NETBOX_TENANT_NAME",
        "NMS_BASE_URL",
        "NMS_OAUTH_CLIENT_ID",
        "NMS_OAUTH_CLIENT_SECRET",
        "NMS_OAUTH_TOKEN_ENDPOINT",
        "NAV_BASE_URL",
        "NAV_API_KEY"
    ]
    if source == "nms":
        config_dict.update({"ni_base_url": ni_base_url,
        "ni_user_name": ni_user_name,
        "ni_api_key": ni_api_key})
        required_envs.extend([
            "NI_BASE_URL",
            "NI_USERNAME",
            "NI_USER_API_KEY"
        ])
        
    config_manager = ConfigManager(
        build_function=build_nav_config, config=config, op_parameters=config_dict
    )
    config_manager.set_required_envs(required_envs)
    settings = config_manager.check_config()

    if show_settings:
        print_color(settings, color="blue")

    if source == "nms":
        update_nav_with_ni_nms(dry_run, verbose, settings)
    elif source == "netbox":
        _sync_from_netbox(
            settings=settings,
            nav_organization=nav_organization,
            verbose=verbose,
            netbox_tenant_name=netbox_tenant_name
        )
    else:
        raise ValueError(f"Unsupported source: {source}. Must be 'nms' or 'netbox'.")
        
def _sync_from_netbox(settings, nav_organization, verbose, netbox_tenant_name):
    # Resolve tenant name
    tenant_name = (
        netbox_tenant_name
        if netbox_tenant_name
        else settings.netbox.tenant_name
    )
    netbox_api = NetBoxAPI(settings.netbox.url, settings.netbox.api_token)
    tenant = netbox_api.get_tenant(tenant_name)
    if not tenant:
        raise ValueError(f"Tenant '{tenant_name}' not found in NetBox.")

    devices = netbox_api.get_devices(
        tenant_id=tenant.id, status="active", manufacturer="arista"
    )
    site = netbox_api.get_site(tenant_id=tenant.id)
    nav = nav_server(
        settings.nav.base_url, settings.nav.api_key, verbose=verbose
    )
    # Print verbose information
    if verbose:
        print_color(f"devices={devices}", color="blue")
        print_color(f"organization={nav_organization}", color="blue")
        print_color(f"site={site.name}", color="blue")
        print_color(f"tenant={tenant.name}", color="blue")
        
    try:
        nav_devices = nav.get_all_devices()
    except Exception as e:
        print(f"Error fetching NAV devices: {e}")
        sys.exit(1)

    # Add new location
    new_location = nav.add_location(name=site.name)
    
    for device_ in devices:
        if verbose:
            print_color("="*40, color="blue")
        try:
            new_room = nav.add_room(room=device_.location.name, location=site.name)
            device_info = {
                "management_ip": device_.primary_ip4,
                "roomid": device_.location.name,
                "organizationid": nav_organization,
                "categoryid": device_.role.name,
                "hostname": f"{device_.name}",
                "device_type": device_.device_type.part_number,
                "nav_devices": nav_devices,
            }
            new_device = nav.add_new_device(**device_info)
        except AttributeError as e:
            print(f"Missing attribute in device {device_}: {e}")
        except Exception as e:
            print_color(f"Unexpected error processing device {device_}: {e}", color="red")

@click.command()
@click.option(
    "--nav-base-url", required=False, default=None, help="Base URL for NAV API."
)
@click.option("--nav-api-key", required=False, default=None, help="API key for NAV.")
@click.option(
    "--ni-base-url", required=False, default=None, help="Base URL for NI API."
)
@click.option("--ni-user-name", required=False, default=None, help="User name for NI.")
@click.option("--ni-api-key", required=False, default=None, help="API key for NI.")
@click.option(
    "--source",
    default="nms",
    required=False,
    show_default=True,
    help="Specify the source system (e.g., nms).",
)
@nms_credentials
def all(
    config,
    source,
    nav_base_url,
    nav_api_key,
    ni_base_url,
    ni_user_name,
    ni_api_key,
    nms_base_url,
    nms_client_id,
    nms_client_secret,
    nms_client_endpoint,
    dry_run,
    show_settings,
    verbose,
):
    """
    Execute both NI and Nav synchronizations sequentially.
    """

    config_dict = {
        "nav_base_url": nav_base_url,
        "nav_api_key": nav_api_key,
        "ni_base_url": ni_base_url,
        "ni_user_name": ni_user_name,
        "ni_api_key": ni_api_key,
        "nms_base_url": nms_base_url,
        "nms_client_id": nms_client_id,
        "nms_client_secret": nms_client_secret,
        "nms_client_endpoint": nms_client_endpoint,
    }

    required_envs = [
        "NMS_BASE_URL",
        "NMS_OAUTH_CLIENT_ID",
        "NMS_OAUTH_CLIENT_SECRET",
        "NMS_OAUTH_TOKEN_ENDPOINT",
        "NAV_BASE_URL",
        "NAV_API_KEY",
        "NI_BASE_URL",
        "NI_USERNAME",
        "NI_USER_API_KEY",
    ]

    config_manager = ConfigManager(
        build_function=build_nav_config, config=config, op_parameters=config_dict
    )
    config_manager.set_required_envs(required_envs)
    settings = config_manager.check_config()

    if show_settings:
        print_color(settings, color="blue")

    error = False
    try:
        print("Updating NI...")
        update_ni(dry_run, verbose, settings)
    except Exception as err:
        error = True
        print("ERROR: Could not update NI")
        print(str(err))

    try:
        print("Updating NAV...")
        update_nav_with_ni_nms(dry_run, verbose, settings)
    except Exception as err:
        error = True
        print("ERROR: Could not update NAV")
        print(str(err))

    if not error:
        print("All updates done!")


def netbox_arguments(func):
    @click.option(
        "--netbox-url", required=False, default=None, help="Base URL for Netbox."
    )
    @click.option(
        "--netbox-api-token",
        required=False,
        default=None,
        help="API token/key for Netbox.",
    )
    @click.option(
        "--netbox-tenant-name",
        required=True,
        default=None,
        help="Tenant name to be synced in Netbox.",
    )
    @click.option(
        "--nms-base-url", required=False, default=None, help="Base URL for NMS."
    )
    @click.option(
        "--nms-client-id", required=False, default=None, help="Client ID for NMS."
    )
    @click.option(
        "--nms-client-secret",
        required=False,
        default=None,
        help="Client secret for NMS.",
    )
    @click.option(
        "--nms-client-endpoint",
        required=False,
        default=None,
        help="Client endpoint for NMS.",
    )
    @cli_flags
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return update_wrapper(wrapper, func)


@click.command()
@click.option(
    "--source",
    default="nms",
    required=False,
    show_default=True,
    help="Specify the source system (e.g., nms).",
)
@netbox_arguments
def netbox(
    config,
    source,
    netbox_url,
    netbox_api_token,
    netbox_tenant_name,
    nms_base_url,
    nms_client_id,
    nms_client_secret,
    nms_client_endpoint,
    dry_run,
    verbose,
    show_settings,
):
    """Perform synchronization between Netbox and a source target (NMS)."""

    config_dict = {
        "netbox_url": netbox_url,
        "netbox_api_token": netbox_api_token,
        "netbox_tenant_name": netbox_tenant_name,
        "nms_base_url": nms_base_url,
        "nms_client_id": nms_client_id,
        "nms_client_secret": nms_client_secret,
        "nms_client_endpoint": nms_client_endpoint,
    }

    required_envs = [
        "NETBOX_URL",
        "NETBOX_API_TOKEN",
        "NETBOX_TENANT_NAME",
        "NMS_BASE_URL",
        "NMS_OAUTH_CLIENT_ID",
        "NMS_OAUTH_CLIENT_SECRET",
        "NMS_OAUTH_TOKEN_ENDPOINT",
    ]

    config_manager = ConfigManager(
        build_function=build_nav_config, config=config, op_parameters=config_dict
    )
    config_manager.set_required_envs(required_envs)
    settings = config_manager.check_config()

    if show_settings:
        print_color(settings, color="blue")

    # Resolve tenant name
    tenant_name = (
        netbox_tenant_name if netbox_tenant_name else settings.netbox.tenant_name
    )

    # Initialize NetBox and NMS clients
    netbox_api = NetBoxAPI(settings.netbox.url, settings.netbox.api_token)
    nms = nms_server(
        nms_client_id=settings.nms.oauth_client_id,
        nms_client_secret=settings.nms.oauth_client_secret,
        nms_token_endpoint=settings.nms.oauth_token_endpoint,
        urlbase=settings.nms.base_url,
        dry_run=dry_run,
    )

    # Fetch NetBox data
    try:
        (
            tenant,
            site,
            netbox_devices,
            netbox_device_types,
        ) = netbox_api.fetch_netbox_data(tenant_name, dry_run=dry_run)
    except Exception as e:
        error_traceback = traceback.format_exc()
        print_color(f"Error fetching NetBox data: {e}", color="red")
        print_color(f"Traceback:\n{error_traceback}", color="red")
        print_color(f"Error occurred at: {traceback.extract_tb(e.__traceback__)[-1]}", color="yellow")
        print_color(f"Reason: {str(e)}", color="yellow")
        sys.exit(1)

    # Fetch devices from NMS
    try:
        nms_devices = nms.get_all_devices()
    except Exception as e:
        error_traceback = traceback.format_exc()
        print_color(f"Error fetching devices from NMS: {e}", color="red")
        print_color(f"Traceback:\n{error_traceback}", color="red")
        print_color(f"Error occurred at: {traceback.extract_tb(e.__traceback__)[-1]}", color="yellow")
        print_color(f"Reason: {str(e)}", color="yellow")
        sys.exit(1)

    # Create missing device types
    try:
        netbox_api.create_device_types_if_none_exist(
            netbox_device_types, nms_devices, dry_run=dry_run, verbose=verbose
        )
    except Exception as e:
        print_color(f"Error creating device types in NetBox: {e}", color="red")
        sys.exit(1)

    # Create missing devices
    try:
        netbox_api.create_device_if_none_exist(
            netbox_devices, nms_devices, tenant, site, dry_run=dry_run, verbose=verbose
        )
    except Exception as e:
        error_traceback = traceback.format_exc()
        print_color(f"Error creating devices in NetBox: {e}", color="red")
        print_color(f"Traceback:\n{error_traceback}", color="red")
        print_color(f"Error occurred at: {traceback.extract_tb(e.__traceback__)[-1]}", color="yellow")
        print_color(f"Reason: {str(e)}", color="yellow")
        sys.exit(1)


# Add commands to the CLI
cli.add_command(ni)
cli.add_command(nav)
cli.add_command(netbox)
cli.add_command(all)


if __name__ == "__main__":
    cli()
