import os
import pynetbox
import traceback
from typing import Optional
from slugify import slugify
from cnaas_sync.utils import remove_prefix_suffix, print_section, print_color
from cnaas_sync.schema import (
    InterfaceModel,
    DeviceModel,
    DeviceTypeModel,
    IpAddressModel,
    DeviceRole,
    DeviceStatusChoices,
    IPAddressStatusChoices,
    DeviceName,
)


class NetBoxAPI:
    def __init__(self, netbox_url: str, api_token: str):
        # Initialize the pynetbox API instance
        self.nb = pynetbox.api(netbox_url, token=api_token)

    def get_tenant(self, tenant_name: str):
        """Fetch tenant ID by name, raise error if tenant not found."""
        tenant = self.nb.tenancy.tenants.get(name=tenant_name)
        if not tenant:
            raise ValueError(f"Tenant '{tenant_name}' not found in NetBox.")
        return tenant

    def get_site(self, tenant_id: str):
        """Fetch the site that belongs to the tenant, create a new site "<tenant>-default" if none exists."""
        sites = self.nb.dcim.sites.filter(tenant_id=tenant_id)
        if sites:
            return list(sites)[0]
        return None

    def get_or_create_site(self, tenant_id: str):
        """Fetch the site that belongs to the tenant, create a new site "<tenant>-default" if none exists."""
        sites = self.nb.dcim.sites.filter(tenant_id=tenant_id)
        if sites:
            return list(sites)[0]

        tenant = self.nb.tenancy.tenants.get(id=tenant_id)
        new_site_name = f"{str(tenant.name).lower()}-default"
        new_site = self.nb.dcim.sites.create({"name": new_site_name, "slug": slugify(new_site_name), "tenant": tenant.id})
        return new_site

    def get_devices_by_tenant(self, tenant_id: str):
        """Fetch all devices associated with the tenant."""
        devices = self.nb.dcim.devices.filter(tenant_id=tenant_id)
        return list(devices)

    def get_devices(self, tenant_id: str, status: str, manufacturer: str):
        """Fetch all active Arista devices associated with the tenant."""
        devices = self.nb.dcim.devices.filter(
            tenant_id=tenant_id, status=status, manufacturer=manufacturer
        )
        return list(devices)

    def get_device_roles(self):
        """Fetch all device types and return them in a list."""
        device_roles = self.nb.dcim.device_roles.all()
        return list(device_roles)

    def get_ipam_prefix_by_parent_tenant(self, tenant_id: str, management_ip: str):
        """Fetch IPAM prefixes associated with the tenant."""
        prefixes = self.nb.ipam.prefixes.filter(
            tenant_id=tenant_id, contains=management_ip
        )
        return list(prefixes)
    
    def get_ip_address(self, tenant_id: str, vrf_id: str, status: str, assigned=True):
        ip_address = self.nb.ipam.ip_addresses.filter(tenant_id=tenant_id, vrf_id=vrf_id, status=status, assigned_to_interface=assigned)
        return list(ip_address)
    
    def get_ip_address_by_interface(self, interface_id: int):
        ip_addresses = self.nb.ipam.ip_addresses.filter(interface_id=interface_id)
        return list(ip_addresses)
    
    def get_ip_address_by_address(self, ip_address, tenant_id):
        ip_obj = self.nb.ipam.ip_addresses.get(address=ip_address, tenant_id=tenant_id)
        return ip_obj

    def get_device_types(self):
        """Fetch all device types and return them in a list."""
        device_types = self.nb.dcim.device_types.all()
        return list(device_types)

    def get_device_types_by_model(self, model: str, manufacturer: str):
        """Fetch all Arista 7280SR-48C6 device types."""
        device_types = self.nb.dcim.device_types.filter(
            model=model, manufacturer=manufacturer
        )
        return list(device_types)

    def get_test_customs(self):
        """Fetch all device types and return them in a list."""
        return list(self.nb.extras.config_contexts.all())

    def create_device_type(
        self, device_type: DeviceTypeModel, dry_run: bool = False, verbose: bool = False
    ):
        """Create and return a device type."""
        new_device = None
        if dry_run:
            print(f"Dry run: Creating device type {device_type.model}")
        else:
            new_device = self.nb.dcim.device_types.create(device_type.model_dump())
            print(f"Created new device type: {new_device}") if verbose else None
        return new_device

    def create_device(self, device: DeviceModel, dry_run: bool = False):
        """Create a new device in NetBox."""
        try:
            if dry_run:
                print(
                    f"Dry run: Creating device {device.name} with type {device.device_type} for tenant {device.tenant} on site {device.site}"
                )
                return device
            new_device = self.nb.dcim.devices.create(device.model_dump())
            return new_device
        except Exception as e:
            print(f"Error creating device: {e}")
            return None

    def create_interface(self, interface: InterfaceModel, dry_run: bool = False):
        """Create a new interface on a device in NetBox."""
        try:
            if dry_run:
                print(
                    f"Dry run: Creating interface {interface.name} on device {interface.device}"
                )
                return interface
            device = self.nb.dcim.devices.get(name=interface.device)
            if not device:
                raise ValueError(f"Device {interface.device} not found")
            interface_dict = interface.model_dump(exclude={"id"})
            new_interface = self.nb.dcim.interfaces.create(interface_dict)
            return new_interface
        except Exception as e:
            print(f"Error creating interface: {e}")
            return None

    def create_ip_address(self, ip_address: IpAddressModel, dry_run: bool = False):
        """Create a new IP address in NetBox."""
        try:
            if dry_run:
                print(
                    f"Dry run: Creating IP address {ip_address.address} for tenant {ip_address.tenant} on interface.id {ip_address.assigned_object_id}"
                )
                return ip_address
            vrf = self.nb.ipam.vrfs.get(name=ip_address.vrf)
            if not vrf:
                raise ValueError(f"VRF {ip_address.vrf} not found")

            tenant = self.nb.tenancy.tenants.get(name=ip_address.tenant)
            if not tenant:
                raise ValueError(f"Tenant {ip_address.tenant} not found")

            ip_address = self.nb.ipam.ip_addresses.create(ip_address.model_dump())
            return ip_address
        except Exception as e:
            print(f"Error creating IP address: {e}")
            return None

    def get_interface_by_name(
        self, interface_name: Optional[str] = None, interface_id: Optional[id] = None, device_name: Optional[str] = None
    ):
        """Fetch interface by name and optionally filter by device name."""
        try:
            if interface_name:
                query_params = {"name": interface_name}
            elif interface_id:
                query_params = {"interface_id": interface_id}
                
            if device_name:
                query_params["device"] = device_name
            return self.nb.dcim.interfaces.get(**query_params)
        except Exception as e:
            print(f"Error retrieving interface: {e}")
            return None
        
    def get_interface_by_id(self, interface_id: Optional[id] = None, device_name: Optional[str] = None
    ):
        """Fetch interface by name and optionally filter by device name."""
        try:
            if interface_id:
                query_params = {"interface_id": interface_id}
                
            if device_name:
                query_params["device"] = device_name
            
            return list(self.nb.dcim.interfaces.filter(**query_params))
        except Exception as e:
            print(f"Error retrieving interface: {e}")
            return None

    def get_vrf_by_tenant(self, tenant):
        """Fetch VRF by tenant."""
        try:
            vrf = self.nb.ipam.vrfs.filter(tenant_id=tenant.id, name__ic=tenant.name)
            vrf_list = list(vrf)
            return vrf_list[0] if len(vrf_list) > 0 else None
        except Exception as e:
            print(f"Error fetching VRF with tenant {tenant}: {e}")
            return None

    def create_device_types_if_none_exist(
        self,
        netbox_device_types,
        nms_devices,
        dry_run: bool = False,
        verbose: bool = False,
    ):
        # Use to keep track of (model, vendor) pair
        seen_pairs = set()

        netbox_device_models = set(
            d.part_number for d in netbox_device_types if d.part_number
        )
        nms_devices_models = set(d["model"] for d in nms_devices)
        type_name_not_in_netbox = nms_devices_models.difference(netbox_device_models)

        # Filter to ensure no two elements have the same (model, vendor) pair
        types_not_in_netbox = filter(
            lambda x: x["model"] in type_name_not_in_netbox
            and not (
                (x["model"], x["vendor"]) in seen_pairs
                or seen_pairs.add((x["model"], x["vendor"]))
            ),
            nms_devices,
        )
        types_not_in_netbox = list(types_not_in_netbox)

        if dry_run or verbose:
            synced_message = "ALL DEVICE TYPES ARE SYNCED IN NETBOX"
            unsynced_message = "DEVICE TYPES NOT FOUND IN NETBOX"

            # Select the appropriate message and color based on the presence of unsynced device types
            section_message = (
                unsynced_message if types_not_in_netbox else synced_message
            )
            section_color = "blue" if types_not_in_netbox else "green"
            print_section(section_message, color=section_color)
            print_color(
                set(d["model"] for d in types_not_in_netbox), color="red"
            ) if types_not_in_netbox else None

        # To create a new device type, we only need the (model, vendor) pairs from NMS
        for dev_type in types_not_in_netbox:
            try:
                # Extract and process model and vendor
                model = dev_type.get("model")
                vendor = dev_type.get("vendor")

                if not model or not vendor:
                    print_color(
                        f"Skipping device type due to missing model or vendor: {dev_type}",
                        color="yellow",
                    )
                    continue

                trimmed_model = remove_prefix_suffix(model)
                dtype = DeviceTypeModel.create(
                    manufacturer=vendor, model=trimmed_model, part_number=model
                )
                self.create_device_type(dtype, dry_run=dry_run)

                print_color(
                    f"Successfully created device type: {trimmed_model} (Vendor: {vendor})",
                    color="green",
                ) if verbose else None

            except Exception as e:
                print_color(
                    f"Failed to create device type for: {dev_type}. Error: {e}",
                    color="red",
                )
                
    def _create_ip_for_interface(self, ip_address, tenant, interface_id, dry_run=False):
        """Create an IP address for an interface."""
        # Check if IP already exists
        existing_ip = self.get_ip_address_by_address(ip_address=ip_address, tenant_id=tenant.id)
        if existing_ip:
            raise ValueError(f"Ip-address {existing_ip} already exist for tenant {tenant}")
            
        tenant_vrf = self.get_vrf_by_tenant(tenant=tenant)
        ip_addr_data = {
            "address": ip_address,
            "vrf": tenant_vrf.name,
            "tenant": tenant.name,
            "status": IPAddressStatusChoices.STATUS_ACTIVE,
            "assigned_object_type": "dcim.interface",
            "assigned_object_id": interface_id,
        }
        ip_addr_model = IpAddressModel.create(**ip_addr_data)
        return self.create_ip_address(ip_addr_model, dry_run=dry_run)
        
    def _create_mgmt_interface(self, device_name, dry_run=False):
        """Create a MGMT interface for a device."""
        interface_data = {
            "device": DeviceName(name=device_name),
            "name": "MGMT",
            "type": "virtual",
        }
        interface_model = InterfaceModel(**interface_data)
        return self.create_interface(interface_model, dry_run=dry_run)
        
        
    def _create_device_helper(self, device, tenant, site, dry_run=False):
        device_data = {
            "name": device["hostname"],
            "role": DeviceRole.from_string(device["device_type"]),
            "site": site.name,
            "type": remove_prefix_suffix(device["model"]),
            "tenant": tenant.name,
            "manufacturer": device["vendor"],
            "status": DeviceStatusChoices.STATUS_STAGED,
            "serial": device["serial"],
        }
        device_model = DeviceModel.create(**device_data)
        return self.create_device(device_model, dry_run=dry_run)



    def _rollback_creation(self, created_objects, error):
        """Roll back created objects in reverse order."""
        print(f"Error occurred: {error}. Rolling back changes...")
        if created_objects.get("ip_address"):
            try:
                self.nb.ipam.ip_addresses.delete(created_objects["ip_address"].id)
            except Exception as e:
                self._log_error(f"Rollback failed for IP: {e}")
                
        if created_objects.get("interface"):
            try:
                self.nb.dcim.interfaces.delete(created_objects["interface"].id)
            except Exception as e:
                self._log_error(f"Rollback failed for interface: {e}")
                
        if created_objects.get("device"):
            try:
                self.nb.dcim.devices.delete(created_objects["device"].id)
            except Exception as e:
                self._log_error(f"Rollback failed for device: {e}")
    
    
    def _create_device_with_mgmt_interface(self, device, tenant, site, create_device=True, dry_run=False):
        """Create a device with its MGMT interface in a transaction-like manner."""
        created_objects = {"device": None, "interface": None, "ip_address": None}
        try:
            # Create device
            if create_device:
                created_objects["device"] = self._create_device_helper(device=device, tenant=tenant, site=site, dry_run=dry_run)

            # Create an interface
            created_interface = self._create_mgmt_interface(device_name=device["hostname"], dry_run=dry_run)
            created_objects["interface"] = created_interface

            # Create ip-address
            
            created_objects["ip_address"] = self._create_ip_for_interface(
                ip_address=device["management_ip"],
                tenant=tenant,
                interface_id=created_interface.id if created_interface else None,
                dry_run=dry_run)
            
            return created_objects
        except Exception as e:
            # Rollback if any step fails
            self._rollback_creation(created_objects)
            raise e
    
    def create_device_if_none_exist(
        self,
        netbox_devices,
        nms_devices,
        tenant,
        site,
        dry_run: bool = False,
        verbose: bool = False,
    ):
        def print_sync_status_message(title, devices):
            color = "red" if devices else "green"
            status = f"{title.upper()} NOT FOUND" if devices else f"ALL {title.upper()} ARE SYNCED"
            print_section(status, color=color)
            if devices:
                print_color([
                    d['hostname'] if isinstance(d, dict) else d 
                    for d in devices
                ], color="red")
                
        def safe_execute(action, devices, create_device=True):
            """
            Safely execute a synchronization action with error handling.
            
            Args:
                action (callable): Method to execute for each item
                devices (list): devices to process
                create_device (bool, optional): Flag for device creation. Defaults to True.
            """
            for device in devices:
                try:
                    action(device, tenant, site, create_device=create_device, dry_run=dry_run)
                except Exception as e:
                    error_msg = (
                        f"Failed to create device {device['hostname']}" 
                        if create_device 
                        else f"Failed to create interface for {device['hostname']}"
                    )
                    print_section(f"{error_msg}: {e}", color="red")
            
        netbox_device_names = {d.name for d in netbox_devices}        
        
        # Identify devices requiring synchronization
        devices_not_in_netbox = [
            dev for dev in nms_devices 
            if dev['hostname'] not in netbox_device_names
        ]
        
        devices_with_unsynced_interfaces = [
            dev for dev in nms_devices 
            if dev["hostname"] in netbox_device_names and
            self.get_interface_by_name(interface_name='MGMT', device_name=dev["hostname"]) is None
        ]

        devices_with_no_ip_address = [
            dev for dev in nms_devices
            if (dev['hostname'] in netbox_device_names and 
                (interface := self.get_interface_by_name(interface_name='MGMT', device_name=dev['hostname'])) is not None and
                not self.get_ip_address_by_interface(interface.id))
        ]
        
        # Print synchronization status if dry_run or verbose is True
        if dry_run or verbose:
            print_sync_status_message("DEVICES", devices_not_in_netbox)
            print_sync_status_message("DEVICES.MGMT.INTERFACES", devices_with_unsynced_interfaces)
            print_sync_status_message("DEVICES.MGMT.IP_ADDRESSES", devices_with_no_ip_address)
            
        if devices_not_in_netbox:
            safe_execute(self._create_device_with_mgmt_interface, devices_not_in_netbox, create_device=True)      

        if devices_with_unsynced_interfaces:
            safe_execute(self._create_device_with_mgmt_interface, devices_with_unsynced_interfaces, create_device=False)
            
        if devices_with_no_ip_address:
            for device in devices_with_no_ip_address:
                try:
                    interface = self.get_interface_by_name(interface_name='MGMT', device_name=device['hostname'])
                    self._create_ip_for_interface(
                            ip_address=device["management_ip"],
                            tenant=tenant,
                            interface_id=interface.id if interface else None,
                            dry_run=dry_run)
                except Exception as e:
                    print_color(f"Failed to create IP {device['management_ip']} for device {device['hostname']}: {e}", color="red")

    def test(self, tenant_name: str):
        """Test function to fetch and display tenant data."""
        try:
            tenant = self.get_tenant(tenant_name)
            site = self.get_or_create_site(tenant.id)
            devices = self.get_devices_by_tenant(tenant.id)
            device_types = self.get_device_types()

            print(f"Tenant ID for '{tenant_name}': {tenant.id}")
            print(f"Site for tenant: {site.name}")
            print(
                f"Devices for tenant '{tenant_name}': {[device.name for device in devices]}"
            )
            print(
                f"Device types: {[device_type.model for device_type in device_types]}"
            )

        except ValueError as e:
            print(e)

    def fetch_netbox_data(
        self, tenant_name, dry_run: bool = False, verbose: bool = False
    ):
        try:
            # Fetch the tenant
            tenant = self.get_tenant(tenant_name)
            if not tenant:
                raise ValueError(f"Tenant '{tenant_name}' not found in NetBox.")

            # Get or create the site for the tenant
            site = self.get_or_create_site(tenant.id)

            # Get devices associated with the tenant
            netbox_devices = self.get_devices_by_tenant(tenant.id)

            # Get all device types
            netbox_device_types = self.get_device_types()

            if dry_run or verbose:
                print_section("Netbox status", color="blue")
                print(f"Tenant='{tenant_name}'")
                print(f"ID['{tenant_name}']: {tenant.id}")
                print(f"Site['{tenant_name}']: {site.name}")
                print(
                    f"Devices['{tenant_name}']: {[device.name for device in netbox_devices]}"
                )
                print(
                    f"Device_types['{tenant_name}']: {[device_type.model for device_type in netbox_device_types]}"
                )

            return tenant, site, netbox_devices, netbox_device_types

        except Exception as e:
            raise RuntimeError(f"Failed to fetch NetBox data: {e}")


if __name__ == "__main__":
    NETBOX_URL = os.getenv("netbox_url")
    API_TOKEN = os.getenv("netbox_api_token")

    # Initialize the NetBoxAPI class with dependency injection
    netbox_api = NetBoxAPI(NETBOX_URL, API_TOKEN)

    # Run the test function
    netbox_api.test(tenant_name="Konstfack")
