import time
import requests
from cnaas_sync.utils import get_number_of_pages, renew_jwt, print_section


class NMS:
    def __init__(
        self,
        nms_client_id,
        nms_client_secret,
        nms_token_endpoint,
        urlbase,
        nms_jwt_token=None,
        verbose=False,
        dry_run=False,
    ):
        self.urlbase = urlbase
        self.nms_oauth_client_id = nms_client_id
        self.nms_oauth_client_secret = nms_client_secret
        self.nms_oauth_token_endpoint = nms_token_endpoint
        self.last_refresh_time = time.time()
        self.verbose = verbose
        self.dry_run = dry_run
        self.jwt_token = None
        self.refresh = True
        if nms_jwt_token:
            self.jwt_token = nms_jwt_token
            self.refresh = False
        else:
            self.jwt_token = self.refresh_jwt_token()

    def _print(self, value):
        if self.verbose:
            print(value)

    def refresh_jwt_token(self):
        try:
            auth_data = {
                "grant_type": "client_credentials",
                "client_id": self.nms_oauth_client_id,
                "client_secret": self.nms_oauth_client_secret,
            }
            auth_response = requests.post(
                self.nms_oauth_token_endpoint,
                data=auth_data,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                timeout=5,
            )

            auth_response.raise_for_status()
            self.jwt_token = auth_response.json()["access_token"]
            self.last_refresh_time = time.time()
            return self.jwt_token
        except Exception as e:
            raise ValueError("Could not find JWT token: {}".format(e))

    @renew_jwt
    def get_all_devices(self):
        headers = {"Authorization": "Bearer " + self.jwt_token}
        res = requests.get(self.urlbase + "/api/v1.0/devices", headers=headers)
        if res.status_code != 200:
            print(f"Failed to fetch devices from NMS! Status Code: {res.status_code}")
            self._print(self)
            return []
        page_size = 1000
        total_count = int(res.headers["X-Total-Count"])
        number_pages = get_number_of_pages(total_count, page_size)
        accumulated_res = []
        # Todo simultaneously call all the page_size concurrently with asyncio/
        for page in range(number_pages):
            res = requests.get(
                self.urlbase
                + f"/api/v1.0/devices?filter[state]=managed&per_page={page_size}&page="
                + str(page),
                headers=headers,
            )
            if "devices" not in res.json()["data"]:
                print(
                    f"Failed to fetch devices from NMS! Status Code[All devices]: {res.status_code}"
                )
                self._print(self)
                return None
            accumulated_res += res.json()["data"]["devices"]
        if self.dry_run or self.verbose:
            print_section("NMS status", color="blue")
            for dev in accumulated_res:
                print(
                    "[ hostname=",
                    dev["hostname"],
                    ", device_type=",
                    dev["device_type"],
                    ", model=",
                    dev["model"],
                    ", serial=",
                    dev["serial"],
                    ", platform=",
                    dev["platform"],
                    ", vendor=",
                    dev["vendor"],
                    " ]",
                )
        return accumulated_res

    @renew_jwt
    def get_device_type(self, hostname):
        headers = {"Authorization": "Bearer " + self.jwt_token}
        device_url = self.urlbase + "/api/v1.0/device/" + hostname
        res = requests.get(device_url, headers=headers)
        self._print(f"[url={device_url}, response={res}")
        if res.status_code != 200:
            print(
                f"Failed to fetch devices from NMS[Device type]! Status Code: {res.status_code}"
            )
            self._print(self)
            return None
        return res.json()["data"]["devices"][0]["device_type"]

    def __str__(self) -> str:
        token = (
            None
            if self.jwt_token is None
            else str(self.jwt_token)[: int(len(str(self.jwt_token)) * 0.10)]
        )
        return f"[urlbase={self.urlbase},\njwt={token},\nnms_id={self.nms_oauth_client_id},\n\
        nms_secret={self.nms_oauth_client_secret},\nnms_endpoint={self.nms_oauth_token_endpoint},\n\
        last_refresh_time={self.last_refresh_time}]"
