import os
import sys
import configparser
from cnaas_sync.utils import print_color


class ConfigManager:
    required_envs = [
        "NMS_BASE_URL",
        "NMS_OAUTH_CLIENT_ID",
        "NMS_OAUTH_CLIENT_SECRET",
        "NMS_OAUTH_TOKEN_ENDPOINT",
        "NI_BASE_URL",
        "NI_USERNAME",
        "NI_USER_API_KEY",
    ]

    def __init__(self, build_function=None, config=None, op_parameters={}):
        self.config = config
        self.build_function = build_function
        self.optional_parameters = op_parameters

    def check_internals(self):
        for attr, value in vars(self).items():
            print("attr=", attr, "value=", value)

    def check_config(self):
        """Check if configuration is available or fallback to environment variables."""
        if not self.config:
            print_color(
                "No configuration file provided. Falling back to environment variables.",
                color="red",
            )
            return self.handle_env_fallback()
        return Settings(self.config)

    def handle_env_fallback(self):
        """Handle the case where environment variables are missing."""
        missing_envs = Settings.check_environment_variables(self.required_envs)
        if missing_envs:
            print_color(
                f"The following required env {missing_envs} are not set. Falling back to op. parameters.",
                color="red",
            )
            # if not self.config and missing_envs and not self.are_parameters_set():
            print("op.parameters=>", self.optional_parameters)
            if not self.are_parameters_set() and missing_envs:
                print_color(
                    "No configuration file, environment variables, or required credentials provided. Cannot proceed.",
                    color="red",
                )
                sys.exit(1)
            return self.build_and_set_config()

        return Settings()  # Use default settings if no required envs are missing

    def are_parameters_set(self):
        # Sending the values to the function as positional arguments
        return Settings.is_all_parameter_set(*self.optional_parameters.values())

    def build_and_set_config(self):
        """Build configuration from provided parameters."""
        ni_config = self.build_function(*self.optional_parameters.values())
        return Settings(config=ni_config)

    def set_required_envs(self, new_envs):
        self.required_envs = new_envs

    def reset_required_envs(self):
        """Reset `required_envs` to the original predefined list."""
        self.required_envs = [
            "NMS_BASE_URL",
            "NMS_OAUTH_CLIENT_ID",
            "NMS_OAUTH_CLIENT_SECRET",
            "NMS_OAUTH_TOKEN_ENDPOINT",
            "NI_BASE_URL",
            "NI_USERNAME",
            "NI_USER_API_KEY",
        ]


class ConfigSettingsBuilder:
    """
    A builder class for manually managing and creating configuration settings, used for settings
    confgi to the Settings class below.
    Provides methods to add sections, set key-value pairs, and retrieve the configuration object.
    """

    def __init__(self):
        self.config = configparser.ConfigParser()

    def add_section(self, section_name):
        if self.config.has_section(section_name):
            raise ValueError(f"Section '{section_name}' already exists.")
        self.config.add_section(section_name)

    def set_value(self, section_name, key, value):
        if not self.config.has_section(section_name):
            raise ValueError(
                f"Section '{section_name}' does not exist. Add the section first."
            )
        # print("section_name=",section_name, "key=", key, "value=", value)
        self.config.set(section_name, key, value)

    def get_value(self, section_name, key):
        if not self.config.has_section(section_name):
            raise ValueError(f"Section '{section_name}' does not exist.")
        if not self.config.has_option(section_name, key):
            raise ValueError(f"Key '{key}' does not exist in section '{section_name}'.")
        return self.config.get(section_name, key)

    def get_config(self):
        return self.config

    def save_to_file(self, file_path):
        with open(file_path, "w") as configfile:
            self.config.write(configfile)

    def load_from_file(self, file_path):
        try:
            with open(file_path, "r") as configfile:
                self.config.read_file(configfile)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Configuration file '{file_path}' not found."
            ) from e
        except configparser.Error as e:
            raise configparser.Error(
                f"Error parsing configuration file '{file_path}'."
            ) from e


class BaseSettings:
    def __init__(self, config=None):
        self.config = config

    def load_config(self, prefix):
        """Load settings from environment variables or INI file"""
        if self.config:
            self.load_from_ini(prefix)
        else:
            self.load_from_env(prefix)

    def load_from_env(self, prefix):
        """Load settings from environment variables with a given prefix."""
        for attr in vars(self):
            if attr == "config":
                continue
            env_var = f"{prefix}_{attr.upper()}"
            setattr(self, attr, os.getenv(env_var, getattr(self, attr)))

    def load_from_ini(self, section):
        """Load settings from a given section in the INI config."""
        if not self.config or not self.config.has_section(section):
            return
        for attr in vars(self):
            if attr == "config":
                continue

            key = f"{section}_{attr.upper()}"
            setattr(
                self,
                attr,
                self.config.get(section, key, fallback=getattr(self, attr)),
            )

    def assert_settings_valid(self):
        """Ensure all required settings are set."""
        cls_name = type(self).__name__.upper()
        for attr, value in vars(self).items():
            if attr == "config":
                continue

            if not value:  # Check if the attribute is unset or falsy
                env_key = f"{cls_name}_{attr.upper()}"
                raise ValueError(
                    f"Missing required setting: {env_key}. Please update the config file or set the environment variable."
                )
        return True

    def to_dict(self):
        data = {}
        for attr, value in vars(self).items():
            if attr == "config":
                continue
            data[attr] = value
        return data

    def __str__(self):
        cls_name = type(self).__name__
        str_repr = f"{cls_name}: {{\n"
        sensitive_keywords = {"key", "token", "secret"}

        for attr, value in vars(self).items():
            # Skip "config" attribute
            if attr == "config":
                continue

            # Mask sensitive attributes
            if any(keyword in attr.lower() for keyword in sensitive_keywords):
                display_value = "***" if value else None
            else:
                display_value = value

            str_repr += f"\t{attr}={display_value}\n"

        str_repr += "}"
        return str_repr


class NMS(BaseSettings):
    def __init__(self, config=None):
        super().__init__(config)
        self.base_url = None
        self.oauth_client_id = None
        self.oauth_client_secret = None
        self.oauth_token_endpoint = None

    def load(self):
        self.load_config("NMS")


class NetBox(BaseSettings):
    def __init__(self, config=None):
        super().__init__(config)
        self.url = None
        self.api_token = None
        self.tenant_name = "Norpan"

    def load(self):
        self.load_config("NETBOX")


class NI(BaseSettings):
    def __init__(self, config=None):
        super().__init__(config)
        self.base_url = None
        self.username = None
        self.user_api_key = None

    def load(self):
        self.load_config("NI")


class NAV(BaseSettings):
    def __init__(self, config=None):
        super().__init__(config)
        self.base_url = None
        self.api_key = None

    def load(self):
        self.load_config("NAV")


class Settings:
    def __init__(self, config_file=None, config=None):
        self.config_file = os.getenv("config_ini", config_file)
        self.config = config if config else self.load_config()

        self.nms = NMS(self.config)
        self.netbox = NetBox(self.config)
        self.ni = NI(self.config)
        self.nav = NAV(self.config)

        # Load settings for each subsystem
        self.nms.load()
        self.netbox.load()
        self.ni.load()
        self.nav.load()

    def load_config(self):
        """Load the configuration file if it exists."""
        if self.config_file and os.path.exists(self.config_file):
            config = configparser.ConfigParser()
            config.read(self.config_file)
            return config
        return None

    @staticmethod
    def check_environment_variables(env_variables, simulated_env=None):
        """
        Args:
            env_variables (list): A list of required environment variable names.
            simulated_env (dict, optional): A dictionary simulating the environment
                                            for testing purposes. Defaults to None.
        Returns: boolean
        """

        environment = simulated_env if simulated_env else os.environ

        results = {}
        envs_not_set = set()
        for var in env_variables:
            results[var] = var in environment and bool(environment[var].strip())
            if not results[var]:
                envs_not_set.add(var)

        return envs_not_set

    @staticmethod
    def is_all_parameter_set(*args):
        """
        Return True if all arguments are set, otherwise False.
        """
        return not any(not arg for arg in args)

    def __str__(self):
        return f"Settings(\n {self.nms},\n{self.netbox},\n{self.ni},\n{self.nav})\n"
